/****************************************************************************
**
** Copyright (C) Datafox GmbH.
** All rights reserved.
** Contact: Datafox GmbH (develop@datafox.de)
**
** This file is part of the DFComDLL module of the Datafox SDK.
**
****************************************************************************/
#pragma once
#ifndef CHANNEL_BRIDGE_H
#define CHANNEL_BRIDGE_H

// Basisklassen
#include "log/df_log.h"

// Klassen fuer Multimasterprotokolle.
#include "channel_1_2.h" // Version 1/2
#include "channel_4.h"   // Version 4

#include <map>
#include <thread>
#include <mutex>
#include <string>

struct StructBindInfo {
    int state;
    int typeDevice;
    unsigned int serial;
    std::string ip;
};

struct StructBindingInfo {
    int channelId;
    StructBindInfo bi;
};

class ChannelBridge : public CDFLog
{
public:
    // Konstruktor fuer Kanalinitialisierung
    ChannelBridge( int connectionId );
    virtual ~ChannelBridge();

    // Setzt die Kanalnummer
    inline void setConnectionId( int connectionId )
    {
        // Zugewiesene Kanalnummer
        m_connection_id = connectionId;
        m_log_check.m_connection_id = connectionId;

        // Kanalnummer setzen
        channel4_.set_connection_id( connectionId );
        // Kommunikationsbasis von _channel4 an andere uebergeben
        channel1_2_.SetChannelBase( channel4_.get_connection() );
    }

    // Setzt Klassenzustand
    inline void setActiveModeState(Datafox::Communication::active_mode_state as)
    {
        channel4_.set_active_mode_state(as);
    }
    // Prueft ob Klasse bereits "in Verwendung"
    inline Datafox::Communication::active_mode_state activeModeState()
    {
        return channel4_.get_active_mode_state();
    }
    // Gibt direkt zurück ob der Kanal Aktiv ist
    inline bool isActive()
    {
        return (activeModeState() == Datafox::Communication::ams_active);
    }

    StructBindInfo bindInfo_;
    // Setzt die Information ob die Kanalklasse an ein bestimmtes Geraet gebunden ist.
    inline void setBind(StructBindInfo bi)
    {
        bindInfo_ = bi;
    }
    // Prueft ob Klasse alle Informationen fuer eine Kanalbindung hat.
    inline int isBind()
    {
        return bindInfo_.state;
    }

    // Verbindungsannahme von uebergebenem Listen-Socket
    int connect(SOCKET *listenSocket, int connectionId, unsigned char deviceAddress = Protocol_base::dv_device_address_stand_alone)
    {
        return channel4_.connect(listenSocket, connectionId, deviceAddress);
    }

    // Setzen der CallBack-Funktion fuer Datensatzbenachrichtigung
    void setCallBackNotifyRecord(int (*CBNotifyRecord)(CNotifyRecord&))
    {
        channel4_.set_call_back_notify_record_available( CBNotifyRecord );
    }

    // Main - Funktion fuer Befehlsausfuehrung in Workerthread
protected:

    //#################################
    // CallRoutinen fuer die Abarbeitung anstehender Kommunikationsbefehle
    //	virtual int CallFunction(short module, short func, unsigned char *data, int &count);
    // NotifyRoutine fuer die Abarbeitung von Benachrichtigungen unter den Klassen
    //	int CallFunctionNotify(short module, short func, unsigned char *data, int &count)
    //	{ return _channel4.CallFunctionNotify(module, func, data, count); }

    // Für das Projekt Spar bei Freudenberg-Id.
    // Da die DLL mit JNA eingebunden wurde ist eine mehrfachansteuerung von mehreren Threads abzufangen.
    std::mutex mutexAPI_;

public:
    std::thread::id threadIdOpen_;

    bool tryEnterCriticalSection();
    void enterCriticalSection();
    void leaveCriticalSection();

public:
    inline void setLastErrorNumber(unsigned char deviceAddress, int errorNumber)
    {
        if (deviceAddress == 255)
        {
            for (auto iter = mapLastErrorNumber_.begin(); iter != mapLastErrorNumber_.end(); ++iter)
                mapLastErrorNumber_[iter->first] = errorNumber;
        }
        else
        {
            mapLastErrorNumber_[deviceAddress] = errorNumber;
        }
    }
    inline int getLastErrorNumber(unsigned char deviceAddress)
    {
        return mapLastErrorNumber_[deviceAddress];
    }
private:
    std::map<unsigned char, int> mapLastErrorNumber_;

public:
    /////////////////////////
    // Schnittstelle

    bool isOpen();
    int open(int deviceAddress, int iType, const std::string &szComString, int iComValue, int iComTimeOut);
    int openAEIII(int iType, const char *szComString, int iComValue, int iComTimeOut);
    void close(void);

    /// Ermitteln der uebergebenen Parameter an open.
    inline void getOpenParams( int &comm_type, std::string &comm_string, uint32_t &comm_value, uint32_t &comm_timeout ) 
    {
        channel4_.get_connection().get_open_parameters( (Connection_base::Connection_type&)comm_type, comm_string, comm_value, comm_timeout );
    };

    // Geraeteadresse pruefen.
    inline bool isValidDeviceAddress(int deviceAddress)
    {
        return (deviceAddress > 0 && deviceAddress < 255);
    }
    // Busnummer des Kommunikationspartners
    unsigned char setDeviceAddressReceiver(unsigned char deviceAddress); 
    unsigned char deviceAddressReceiver();

    // Busnummer der DLL - Klassen
    unsigned char setDeviceSender(unsigned char deviceAddress); 
    unsigned char deviceAddressSender();

    // Schnittstelle nach Fehler bereinigen
    void resetWithError(int err);

    // Speicher bereinigen
    void reset();

public:
    /////////////////////////
    // Kommunikationshelper

    int read(uint8_t *buffer, std::size_t receiveCount, std::size_t &receivedCount);
    int write(const uint8_t *buffer, std::size_t transmitCount, std::size_t &transmittedCount);

    void setCallBackFunction(int (*pFunction)(void));
    int getProtocolVersion2(unsigned char deviceAddress, unsigned char detailed, int *piError);
    // Gibt die ermittelte Protokollversion zurueck
    unsigned char getProtocolVersion(unsigned char deviceAddress);
    unsigned char setProtocolVersion(unsigned char deviceAddress, unsigned char version);
    unsigned char protocolVersion_[251];
    bool preferAEIII_;     // Bevorzugt auf AEIII+ pruefen


    /////////////////////////
    // Zugriff auf HANDLE/SOCKET

    /// Ermitteln des Schnittstellenhandle.
    HANDLE getComPort();
    /// Setzen der Schnittstellenparameter.
    int setComPort(HANDLE hCom, int iType, int iComValue, int iTimeOut);

    int comReleaseHandle();
    int comUseHandle(HANDLE h);


    /////////////////////////
    // Helper

    /////////////////////////
    // Befehle

    // Allgeimein

    // Anzahl der letzten Ermittlungsversuche halten.
    int devicePollRetry_;
    // Geraeteerreichbarkeit pruefen
    int checkDevice(unsigned char deviceAddress, int *piError, int iPollRetry);

    // PZE Master IV warmstart.
    int restart(unsigned char deviceAddress, int *piError);
    // Reset des Gerätes
    int reset(unsigned char deviceAddress, int mode, int *piError);

    // Senden einer Bestaetigungsmeldung an das AEIII+
    int sendMessage(unsigned char deviceAddress, unsigned char ucDelayTime,
        unsigned char ucAction,	unsigned char ucAudio, const std::string &text, int *piError);

    // Senden eines Infotextes fuer Geraetestatus.
    int sendInfotext(unsigned char deviceAddress, const std::string &text, int *piError);

    // RTC - Wert schreiben / lesen
    int setDateTime(unsigned char deviceAddress, const unsigned char *pucDateTime, int *piError);
    int getDateTime(unsigned char deviceAddress, unsigned char *pucDateTime, int *piError);

    // Globale-/Systemvariable schreiben / lesen
    int setGlobVar(unsigned char deviceAddress, const char *name, int iType, const char *val, int *piError);
    int getGlobVar(unsigned char deviceAddress, const char *name, int iType, char *val, int iValueLength, int *piError);

    // Relais setzen/ruecksetzen und zustand abfragen.
    int closeRelay(unsigned char deviceAddress, int iNum, int iTime, int *piError);
    int openRelay(unsigned char deviceAddress, int iNum, int *piError);
    int getRelayState(unsigned char deviceAddress, int iNum, int *piState, int *piTime, int *piError);

    // Seriennummer auslesen
    int getSerialNum(unsigned char deviceAddress, int *piSerialNum, int *piError);
    // Firmwareversion aus Dateiheader lesen
    int getVersionFirmwareFromFile(const std::string& filePath, std::string& version, int *piError);

    /////////////////////////
    // System Device Log
    // Datensatz lesen
    int systemReadRecord( unsigned char deviceAddress, uint8_t *buffer, std::size_t &size, int32_t &error );
    // Datensatz quittieren.
    int systemQuitRecord(unsigned char deviceAddress, int *error);
    // Datensatzzeiger ruecksetzen
    int systemRestoreRecords(unsigned char deviceAddress, int *piError);

    // Aktuellen Status des Flash abrufen
    int getFlashStatus(unsigned char deviceAddress, int *piStatus, int *piError);
public:
    /////////////////////////
    // Datensatz- / Listenbeschreibungen

    int getTableDescription(unsigned char deviceAddress, unsigned char ucList, int *piError);
protected:
    //int getTables4(unsigned char deviceAddress, char *typeName);
    int getListName(unsigned char deviceAddress, unsigned char ucList, std::string &name, int *piError);


public:
    /////////////////////////
    // Datensaetze

    // Ermitteln des ersten Datensatzes
    int collectData(unsigned char deviceAddress, int *piError);
    // Ermitteln des naechsten Datensatzes
    int getDatensatz( uint8_t *buffer, std::size_t &size, int32_t &error );
    // Pruefen ob min. ein Datensatz im AEIII+ vorhanden.
    int checkData(unsigned char deviceAddress, int *piError);
    // Loescht alle Datensaetzte im AEIII+
    int clearData(unsigned char deviceAddress, int *piError);
    // Datensatzzeiger ruecksetzen
    int restoreRecords(unsigned char deviceAddress, int *piError);
    // Datensatz lesen
    int readRecord(unsigned char deviceAddress, uint8_t *buffer, std::size_t &size, int32_t &error);
    // Datensatz quittieren.
    int quitRecord(unsigned char deviceAddress, int *error);
    // Anzahl der Datensaetze oder verwendetes Speichervolumen.
    int recordVolume(unsigned char deviceAddress,  int &number, int &memoryUsage, int &memorySize);
protected:
    // Datensatzfunktionen, Variable und Puffer fuer letzten gelesenen.
    std::size_t m_record_size;
    int m_record_error;
    uint8_t m_record_data[MAX_MB_PAKET];


public:
    /////////////////////////
    // Listen

    // Laden der Importierten Listendaten
    int writeListData(unsigned char deviceAddress, int *piError);
protected:
    int writeListData4(unsigned char deviceAddress);
    //int DownloadLists(unsigned char deviceAddress, int *piError);

public:
    /////////////////////////
    // Zutrittslisten (Version 1 / Version 2)

    // Laden der Importierten Listendaten fuer Zutrittskontrolle
    int writeEntranceListData(unsigned char deviceAddress, int iListNum, int *piError);
    int writeEntrance2ListData(unsigned char deviceAddress, int iListNum, int *piError);
protected:
    int writeEntrance2ListData4(unsigned char deviceAddress, unsigned char listNum);
    int writeEntrance2List4(unsigned char deviceAddress, unsigned char listNum);


public:
    /////////////////////////
    // Timeboylisten

    // Laden der Importierten Listendaten fuer Timeboy
    int writeTimeboyListData(unsigned char deviceAddress, unsigned char id, unsigned char num, int reserve, int *piError);
protected:
    int writeTimeboyListData4(unsigned char deviceAddress, unsigned char id, unsigned char num, int reserve);
    int writeTimeboyList4(unsigned char deviceAddress, unsigned char id, unsigned char num, int reserve);


public:
    /////////////////////////
    // Setup

    int writeStudioFile(unsigned char deviceAddress, const std::string &filePath, int *piError);
    int readStudioFile(unsigned char deviceAddress, const std::string &filePath, int *piError);

    // Veraenderungen an der Setupdatei vornehmen
    int modifyStudioFile(const std::string &sourceFile, const std::string &destFile, const std::string &changeData, int flags, int *piError);

protected:
    int writeSetupData(int *piError);
    int writeSetupData4(unsigned char deviceAddress);
    int writeSetupData4_2(unsigned char deviceAddress);
    int readSetupData(const std::string &filePath, int *piError);
    int readSetupData4(unsigned char deviceAddress);
    int readSetupData4_2(unsigned char deviceAddress);

public:
    /////////////////////////
    // Firmware
    int writeFileDevice(unsigned char deviceAddress, std::string &filePath, int *piError);
protected:
    int writeDevice4(unsigned char deviceAddress);

public:
    /////////////////////////
    // Module
    int writeModule(unsigned char deviceAddress, int type, const std::string &filePath, int *piError);
    int writeTextData(unsigned char deviceAddress, const std::string &filePath);
    int writeTouchKeyData(unsigned char deviceAddress, const std::string &filePath);
    int writeDisplayColors(unsigned char deviceAddress, const std::string &filePath);
    int writeWlanConfigData(unsigned char deviceAddress, const std::string &filePath);
    int readWlanConfigData(unsigned char deviceAddress, const std::string &filePath);
    int writeHardwareConfigData(unsigned char deviceAddress, const std::string &filePath);
    int readHardwareConfigData(unsigned char deviceAddress, const std::string &filePath);
    int writeFileData(unsigned char deviceAddress, const std::string &filePath);
    int readFileData(unsigned char deviceAddress, const std::string &filePath);

public:
    /////////////////////////
    // Log / Trace

    /// Setzen des Logdateinamen.
    void setLogFileName(const std::string &szLogFileName);

    /// Log anschalten.
    inline void setLogOn()  
    { channel4_.set_trace_mode(); }

    /// Log ausschalten.
    inline void setLogOff() 
    { channel4_.set_trace_mode( Trace_base::Trace_mode::off ); }

    void setTraceFileName( const std::string &filePath )
    {
        channel4_.set_trace_file_path( filePath );
    }
	 void change_trace_file_path( const std::string& file_path );

    /////////////////////////
    // Passwort
    void setPassword(const char *password)
    {
        channel4_.set_password_communication(password);
    }
public:
    int writeMemToFile(unsigned char deviceAddress, int type, int address, int size, const std::string &filePath);
    int writeFileToMem(unsigned char deviceAddress, const std::string &filePath);

public:	
    /////////////////////////
    // Kanalklassen fuer Protokollverwendung.
    CChannel1_2 channel1_2_;
    Channel4 channel4_;

    inline int resultToError(int iRes, int *piError)
    {
        if (iRes < 0)
        {
            // Kommunikationsfehler
            iRes = -iRes;
            // Fehlercode umsetzen
            *piError = iRes;
            iRes = 0;
        }
        else if (iRes > 0) 
        {
            // Wenn der Modulfehler -Funktion unbekannt- ist,
            // dann ist das Geraet zwar aktiv der Befehl
            // wird jedoch nicht unterstuetzt.
            if ( ((iRes >> 8) & 0xFF) == 2 )
            {
                *piError = IV_ERROR_FUNCTION_UNAVAILABLE;
            }
            else if ( ((iRes >> 8) & 0xFF) == 16 ) 
            {
                // Angesprochener Slot ist nicht vorhanden.
                *piError = IV_ERROR_SLOT_NONE;
            }
            else if ( ((iRes >> 8) & 0xFF) == 17 ) 
            {
                // Kein Timeboy im Slot vorhanden.
                *piError = IV_ERROR_NO_TIMEBOY;
            }
            else if ( ((iRes >> 8) & 0xFF) == 18 ) 
            {
                // Timeboytyp noch nicht erkannt.
                *piError = IV_ERROR_TIMEBOY_NOT_READY;
            }
            else if ( ((iRes >> 8) & 0xFF) == 19 ) 
            {
                // Es wird versucht ein TimeboyI-III anzusprechen.
                *piError = IV_ERROR_TIMEBOY_I_II_III;
            }
            else if ( ((iRes >> 8) & 0xFF) == 20 ) 
            {
                // Ausführung konnte wegen belegtem Kommunikationskanal nicht weitergereicht werden.
                *piError = IV_ERROR_DOCKING_BUSY;
            }
            else if ( (iRes & 0xFF) == 28 ) 
            {
                // Passwort fuer Kommunikation wird benoetigt.
                *piError = IV_ERROR_NEED_PASSWORD;
            }
            else if ( ( iRes & 0xFF ) == 29 )
            {
                // Nach 3 maliger Passwortfalscheingabe ist Geraet fuer 60 Sekunden gesperrt.
                *piError = IV_ERROR_PASSWORD_LOCK;
            }
            else
            {
                // Funktionsfehler
                *piError = IV_ERROR_PROCESSING;
            }

            // Fehler
            iRes = 0;
        } else {

            // OK
            *piError = IV_ERROR_SUCCESS;
            iRes = 1;
        }

        return iRes;
    }
public:
    /////////////////////////
    // Schnittstelle zu Channel1/Channel2
    int getModus(unsigned char deviceAddress, int *error);

    /////////////////////////
    // Schnittstelle zu Channel4
    void setCallBackFunctionWithParams(int (*pFunction)(unsigned int, unsigned int));
    unsigned int crc32(unsigned char *data, unsigned int size, unsigned int crc = 0xFFFFFFFF, bool closure = true);

    // MODULE_HARDWARE

    // Einfache Anfrage ob das Geraet antwortet.
    int ping(unsigned char deviceAddress);

    // Typ lesen
    int setTypeDevice(unsigned char deviceAddress, unsigned char type);
    int getTypeDevice(unsigned char deviceAddress, unsigned char &type);
    int setTypeBoard(unsigned char deviceAddress, unsigned char type);
    int getTypeBoard(unsigned char deviceAddress, unsigned char &type);
    int setTypeTCPIP(unsigned char deviceAddress, unsigned char val);
    int getTypeTCPIP(unsigned char deviceAddress, unsigned char &val);
    int setTypeModemMC(unsigned char deviceAddress, unsigned char val);
    int getTypeModemMC(unsigned char deviceAddress, unsigned char &val);
    int setTypeDisplay(unsigned char deviceAddress, unsigned char type);
    int getTypeDisplay(unsigned char deviceAddress, unsigned char &type);
    int setTypeKeypad(unsigned char deviceAddress, unsigned char type);
    int getTypeKeypad(unsigned char deviceAddress, unsigned char &type);
    int setTypeTransponder(unsigned char deviceAddress, unsigned char type);
    int getTypeTransponder(unsigned char deviceAddress, unsigned char &type);
    int setTypeMMC(unsigned char deviceAddress, unsigned char type);
    int getTypeMMC(unsigned char deviceAddress, unsigned char &type);
    int setTypeRadio(unsigned char deviceAddress, unsigned char type);
    int getTypeRadio(unsigned char deviceAddress, unsigned char &type);
    int setTypeGPS(unsigned char deviceAddress, unsigned char val);
    int getTypeGPS(unsigned char deviceAddress, unsigned char &type);
    int setTypeDCF77(unsigned char deviceAddress, unsigned char type);
    int getTypeDCF77(unsigned char deviceAddress, unsigned char &type);
    int setTypeFingerprint(unsigned char deviceAddress, unsigned char type);
    int getTypeFingerprint(unsigned char deviceAddress, unsigned char &type);
    int setTypeExternalReader(unsigned char deviceAddress, unsigned char val);
    int getTypeExternalReader(unsigned char deviceAddress, unsigned char &val);
    int setTypeTouch(unsigned char deviceAddress, unsigned char val);
    int getTypeTouch(unsigned char deviceAddress, unsigned char &val);
    int setTypeSensor(unsigned char deviceAddress, unsigned char val);
    int getTypeSensor(unsigned char deviceAddress, unsigned char &val);

    // Komponente lesen
    int setComponentRS485_0(unsigned char deviceAddress, unsigned char val);
    int getComponentRS485_0(unsigned char deviceAddress, unsigned char &val);
    int setComponentRS485_1(unsigned char deviceAddress, unsigned char val);
    int getComponentRS485_1(unsigned char deviceAddress, unsigned char &val);
    int setComponentUSB(unsigned char deviceAddress, unsigned char val);
    int getComponentUSB(unsigned char deviceAddress, unsigned char &val);
    int setComponentInAnalog(unsigned char deviceAddress, unsigned int val);
    int getComponentInAnalog(unsigned char deviceAddress, unsigned int &val);
    int setComponentInDigital(unsigned char deviceAddress, unsigned int val);
    int getComponentInDigital(unsigned char deviceAddress, unsigned int &val);
    int setComponentOutDigital(unsigned char deviceAddress, unsigned int val);
    int getComponentOutDigital(unsigned char deviceAddress, unsigned int &val);
    int setComponentBattery(unsigned char deviceAddress, unsigned char val);
    int getComponentBattery(unsigned char deviceAddress, unsigned char &val);
    int setComponentCharge(unsigned char deviceAddress, unsigned char val);
    int getComponentCharge(unsigned char deviceAddress, unsigned char &val);

    // Option schreiben/lesen
    int setOptionIOModule(unsigned char deviceAddress, unsigned char val);
    int getOptionIOModule(unsigned char deviceAddress, unsigned char &val);
    int setOptionTBAdaption(unsigned char deviceAddress, unsigned char val);
    int getOptionTBAdaption(unsigned char deviceAddress, unsigned char &val);
    int setOptionComServer(unsigned char deviceAddress, unsigned char val);
    int getOptionComServer(unsigned char deviceAddress, unsigned char &val);
    int setOptionTPWrite(unsigned char deviceAddress, unsigned char val);
    int getOptionTPWrite(unsigned char deviceAddress, unsigned char &val);
    int setOptionHTTP(unsigned char deviceAddress, unsigned char val);
    int getOptionHTTP(unsigned char deviceAddress, unsigned char &val);
    int setOptionLanguage(unsigned char deviceAddress, unsigned char val);
    int getOptionLanguage(unsigned char deviceAddress, unsigned char &val);
    int setOptionBiokeyUpdate(unsigned char deviceAddress, unsigned char val);
    int getOptionBiokeyUpdate(unsigned char deviceAddress, unsigned char &val);
    int setSetupOptions(unsigned char deviceAddress, unsigned short mask, unsigned short options);
    int getSetupOptions(unsigned char deviceAddress, unsigned short &mask, unsigned short &options);
    int setOptionFirmware(unsigned char deviceAddress, int mask, int options);
    int getOptionFirmware(unsigned char deviceAddress, int &mask, int &options);
    int setPayableOptions(unsigned char deviceAddress, unsigned char mask, unsigned char options);
    int getPayableOptions(unsigned char deviceAddress, unsigned char &mask, unsigned char &options);



    // HWConfig lesen
    int setHWConfig(unsigned char deviceAddress, unsigned char val);
    int getHWConfig(unsigned char deviceAddress, unsigned char &val);

    // Gesamte Hardwareinfo lesen
    int getHardwareInfo(unsigned char deviceAddress, StructHardwareInfo &hwInfo);
    // Gesamte Hardwareinfo lesen
    int getHardwareInfo( unsigned char deviceAddress, HardwareInfo &hardwareInfo );
    // Gesamte Hardwareinfo in String umsetzen.
    int getInfos(StructHardwareInfo hwInfo, std::string &info);


    // MODULE_SYSTEM

    // Systemreset; Speicherbereiche loeschen
    //	int SystemReset(unsigned char deviceAddress, unsigned short modus);

    // Firmwareversion lesen
    int getVersionFirmware(unsigned char deviceAddress, std::string &version);
    int getVersionBootloader(unsigned char deviceAddress, std::string &version);

    // Seriennummer 
    int setSerialNumber(unsigned char deviceAddress, unsigned int serial);
    int getSerialNumber(unsigned char deviceAddress, unsigned int &serial);

    // Datum/Uhrzeit setzen/lesen
    //	int SetRTC(unsigned char deviceAddress, char *time);
    //	int GetRTC(unsigned char deviceAddress, char *time);

    // Systemvariable setzen/lesen
    //	int SetSystemVariable(unsigned char deviceAddress, char *name, char *val);
    //	int GetSystemVariable(unsigned char deviceAddress, char *name, char *val);

    // Relais setzen/lesen
    //	int SetRelay(unsigned char deviceAddress, char *name, unsigned char busNum, unsigned char relayNum, unsigned char status, unsigned short time);
    //	int GetRelay(unsigned char deviceAddress, char *name, unsigned char busNum, unsigned char relayNum, unsigned char &status, unsigned short &time);

    // Datensegmentuebertragung beginnen/enden
    int beginSegment(unsigned char deviceAddress, unsigned short art, unsigned int &address, unsigned int size, unsigned int crc);
    int endSegment(unsigned char deviceAddress, unsigned short art, unsigned int crc);

    // Flash setzen/lesen (die ersten 16kByte Bootloader koennen nur gelesen werden)
    int setFlash(unsigned char deviceAddress, unsigned int address, unsigned char *buf, int count);
    int getFlash(unsigned char deviceAddress, unsigned int address, unsigned char *buf, int count);
    int getFlashPage(unsigned char deviceAddress, std::size_t pageNum, std::size_t offs, uint8_t *buffer, std::size_t &count);

    // EEPROM setzen/lesen
    int setEEPROM(unsigned char deviceAddress, unsigned int address, unsigned char *buf, int count);
    int getEEPROM(unsigned char deviceAddress, unsigned int address, unsigned char *buf, int count);

    // RAM setzen/lesen
    int setRAM(unsigned char deviceAddress, unsigned int address, unsigned char *buf, int count);
    int getRAM(unsigned char deviceAddress, unsigned int address, unsigned char *buf, int count);

    // PROGRAM setzen/(lesen unterbunden)
    //	int SetProgram(unsigned char deviceAddress, unsigned int address, unsigned char *buf, unsigned short count);
    //	int GetProgram(unsigned char deviceAddress, unsigned int address, unsigned char *buf, unsigned short count);

    // MMC setzen/lesen 
    int setMMC(unsigned char deviceAddress, unsigned int address, unsigned char *buf, int count);
    int getMMC(unsigned char deviceAddress, unsigned int address, unsigned char *buf, int count);

    // Systembios aufrufen/verlassen
    int enterBios(unsigned char deviceAddress, unsigned char protocolVersion);
    int exitBios(unsigned char deviceAddress);

    // Interface setzen/lesen (RS232, TCPIP...)
    int setInterface(unsigned char deviceAddress, unsigned char type);
    int getInterface(unsigned char deviceAddress, unsigned char &type);

    // Baud-Rate setzen/lesen
    int setBaud(unsigned char deviceAddress, unsigned int br);
    int getBaud(unsigned char deviceAddress, unsigned int &br);

    // DeviceID setzen/lesen
    int setDeviceIdBios(unsigned char deviceAddress, unsigned short id);
    int getDeviceIdBios(unsigned char deviceAddress, unsigned short &id);

    // Bootloader aktivieren
    int enterBootloader(unsigned char deviceAddress, unsigned char protocolVersion);

    // Monitor Mode vom XPort
    int enterXPortMonitor(unsigned char deviceAddress, char *data, uint16_t &count);
    int enterXPortMonitorExt(unsigned char deviceAddress, char *data, uint16_t &count);
    int commandXPort(unsigned char deviceAddress, char *buf, uint16_t &count);
    int exitXPortMonitor(unsigned char deviceAddress);

    // Oeffnen einer Tabelle
    int tableOpen(unsigned char deviceAddress, const std::string &name, unsigned int options, int &handle, std::string &reserved);
    // Schliessen der Tabelle
    int tableClose(unsigned char deviceAddress, int handle);

    // Setzen eines Filters auf die Tabelle
    int tableSetFilter(unsigned char deviceAddress, int handle, const std::string &filter);
    // Lesen des Filters auf die Tabelle
    int tableGetFilter(unsigned char deviceAddress, int handle, std::string &filter);
    // Loeschen des Filters der Tabelle
    int tableRemoveFilter(unsigned char deviceAddress, int handle);

    // Abfragen der Datensatzanzahl
    int tableGetRowCount(unsigned char deviceAddress, int handle, unsigned int &currentCount, unsigned int &unsortedCount, unsigned int &deletedCount);
    // Abfragen der aktuellen Datensatznummer
    int tableGetCurrentRow(unsigned char deviceAddress, int handle, unsigned int &currentRow);
    // Ueberspringen von n Datensaetzen (offset kann negativ sein), ab top/aktuell/bottom
    int tableSetCurrentRow(unsigned char deviceAddress, int handle, int offset, unsigned short direction);

    // Aendern des aktuellen Datensatzes
    int tableSetCurrentRowData(unsigned char deviceAddress, int handle, unsigned short flags, char separator, const char *data, unsigned short dataSize);
    // Aendern eines Spaltenwerts, des aktuellen Datensatzes
    int tableSetCurrentColumnData(unsigned char deviceAddress, int handle, unsigned short flags, const char *name, const char *data, unsigned short dataSize);
    // Aendern eines Spaltenwerts bei allen aktuell selektieren Zeilen.
    int tableSetAllRowsToColumnData(unsigned char deviceAddress, int handle, unsigned short flags, const char *name, const char *data, unsigned short dataSize);

    // Abfragen des aktuellen Datensatzes
    int tableGetCurrentRowData(unsigned char deviceAddress, int handle, unsigned short flags, char separator, char *data, unsigned short &dataSize);
    // Abfragen eines Spaltenwerts, des aktuellen Datensatzes
    int tableGetCurrentColumnData(unsigned char deviceAddress, int handle, unsigned short flags, const char *name, char *data, unsigned short &dataSize);

    // Daten an die Listentabelle anfuegen.
    int tableAppendRowData(unsigned char deviceAddress, int handle, unsigned short flags, char separator, const char *data, unsigned short dataSize);
    // Loescht den aktuellen Datensatz
    int tableDeleteCurrentRow(unsigned char deviceAddress, int handle, unsigned short flags);
    // Loescht die aktuellen Datensaetze, welche durch den gesetzten Filter verfuegbar sind.
    int tableDeleteAvailableRows(unsigned char deviceAddress, int handle, unsigned short flags);

    // Abfragen einer Modulinformation
    int getInfo(unsigned char deviceAddress, const std::string &name, const std::string &params, std::string &info);

    // Zutrittsbuchung durchfuehren
    int entrance2Identification(unsigned char deviceAddress, const char *TM, const char *Id, const char *Pin, char *status, int &statusLength);
    // LED, Relais auf Tuermodul schalten
    int entrance2OnlineAction(unsigned char deviceAddress, const char *TM, unsigned int mask, unsigned char type, unsigned char duration);

    // Transparentmodus aktivieren/deaktivieren
    int openComServerMode(unsigned char deviceAddress, const char *comm, unsigned int baudRate, unsigned char byteSize, unsigned char parity, unsigned char stopBits, unsigned short flags, unsigned short timeoutToClose);
    int closeComServerMode(unsigned char deviceAddress);

    // Aktive Meldungen aktivieren/deaktieren
    int setNotifyOnDevice(unsigned char deviceAddress, int type, int state);

    // Softwareinformation der Anwendungsfirmware abfragen
    int getSoftwareInfo(unsigned char deviceAddress, unsigned char *swi, int &infoLength);

    // Passwort pruefen
    int passwordCheck(unsigned char deviceAddress, char *password);
    // Passwortschlüssel abrufen
    int getPasswordKey(unsigned char deviceAddress, char *key);

    // Virtuellen Tastendruck ausführen.
    int pressVirtualKey(unsigned char deviceAddress, unsigned char vk, unsigned char flags, int *piError);

    // MODULE_TABLE
#define BIOKEY_OK                                (0x80)  ///<letzer Befehl erfolgreich
#define BIOKEY_ERROR                             (0x81)  ///<unerwarteter Fehler aufgetreten
#define BIOKEY_BAD_IMAGE                         (0x82)  ///<kein oder schlechtes Bild
#define BIOKEY_BAD_FPT                           (0x83)  ///<schlechtes Template, nicht genug Merkmale
#define BIOKEY_DB_FULL                           (0x84)  ///<Datenbank voll
#define BIOKEY_DB_EMPTY                          (0x85)  ///<Datenbank leer
#define BIOKEY_IDS_USED                          (0x86)  ///<übergebene PIDs werden schon benutzt
#define BIOKEY_IDS_UNKNOWN                       (0x87)  ///<übergebene PIDs existieren nicht
#define BIOKEY_RECOGNIZED                        (0x88)  ///<Person wurde erkannt
#define BIOKEY_NOT_RECOGNIZED                    (0x89)  ///<Person wurde nicht erkannt
#define BIOKEY_NO_IMAGE                          (0x8A)  ///<es wurde kein Bild vom Sensor aufgenommen
#define BIOKEY_NO_MATCH                          (0x8B)  ///<kein Treffer
#define BIOKEY_AUTH_FAILED                       (0x8C)  ///<falsches Passwort
#define BIOKEY_FPT_EXISTS                        (0x8D)  ///< an attempt to enroll with message cmd 0x26 failed because a template that matches the newly acquired one was found in the database
#define BIOKEY_FPT_ENROLLED                      (0x8E)  ///< a newly acquired fingerprint has been enrolled successfully
#define BIOKEY_TIMEOUT                           (0x8F)  ///< a timeout occurred during an enroll/identify process
#define BIOKEY_FPT_READ                          (0x90)  ///< a fingerprint was read during “Enroll best of three” procedure
    // Datensatz an Fingerprint-Tabelle anfuegen.
    int fingerprintAppendRecord(unsigned char deviceAddress, unsigned char type, const unsigned char *fingertemplate, int size);
    // Datensatz aus Fingerprint-Tabelle lesen.
    int fingerprintGetRecord(unsigned char deviceAddress, unsigned char &type, int pid, int fid, unsigned char *fingertemplate, int &size);
    // Datensatz aus Fingerprint-Tabelle loeschen.
    int fingerprintDeleteRecord(unsigned char deviceAddressremote, int pid, int fid);
    // Tabelle mit Fingerprintdaten erzeugen
    int fingerprintList(unsigned char deviceAddress, int flags, unsigned int &recordCount);
    // Backup der Fingerprints erstellen
    int fingerprintBackup(unsigned char deviceAddress, const std::string &filePath, int flags);
    // Wiederherstellen der Fingerprints
    int fingerprintRestore(unsigned char deviceAddress, const std::string &filePath, int flags);

    // Systemflags setzen/lesen
    int setSystemFlags(unsigned char deviceAddress, unsigned int flags);
    int getSystemFlags(unsigned char deviceAddress, unsigned int &flags);

    // MODULE_APPLICATION
    // Setzt den Font für die Anzeige der Message und Info - Text
    int setTypeFont(unsigned char deviceAddress, int iType, int *piError);

    // Setzt das Passwort fuer die verschluesselte Kommunikation
    int setCommunicationPassword(const char *password, uint8_t size, uint32_t sessionTimeout, bool global);
    int isGlobalCommunicationPassword();

private:
    int protocolVersionRecordDescription_;
    int protocolVersionListDescription_;

public:
    // Funktionen für Pufferzugriff.
    int GetTableCount(unsigned char ucList = 0);

    int GetTableRecord(unsigned char ucNum, std::string& szName, 
        unsigned char *pucFieldCount, unsigned char *pucVariable = NULL, 
        unsigned char ucList = 0);

    int GetTableField(unsigned char ucTableNum, unsigned char ucFieldNum, 
        std::string& szName, unsigned char *pucType, 
        unsigned char *pucLength, unsigned char ucList = 0);

    void ClearListBuffer();
    int MakeListe(int ListNum, int ListCnt, int ListSize, const unsigned char *Buf);

    void ClearEntranceListBuffer(int iListNum = -1);
    int MakeEntranceList(int ListNum, int ListCnt, int ListSize, const unsigned char *Buf);

    void ClearEntrance2ListBuffer(int iListNum = -1);
    int MakeEntrance2List(int ListNum, int ListCnt, int ListSize, const unsigned char *Buf);

    void ClearTimeboyListBuffer(unsigned char id = 0xFF, unsigned char num = 0xFF);
    int MakeTimeboyList(unsigned char id, unsigned char num, unsigned short recordCount, const unsigned char *buf, unsigned int size, int flags, int *error);

    bool WriteListToBinFile( unsigned char listNum, const char *filePath );
    bool WriteListToTextFile( unsigned char listNum, const char *filePath );
    bool WriteListEntrance2ToBinFile( unsigned char listNum, const char *filePath );
    bool WriteListEntrance2ToTextFile( unsigned char listNum, const char *filePath );
    bool WriteListTimeboyToBinFile( unsigned char listId, unsigned char listNum, const char *filePath );

    std::string GetFileDeviceFileName();
    std::string GetFileDeviceError();
    std::string GetFileDeviceInfo();
    FileDevice &GetFileDevice();

    // Zutrittsbuchung durchfuehren
    int accessControlIdentification(unsigned char deviceAddress, int masterId, const char *TM, const char *Id, const char *Pin, char *status, int &statusLength);
    // LED, Relais auf Tuermodul schalten
    int accessControlOnlineAction(unsigned char deviceAddress, int masterId, const char *TM, unsigned int mask, unsigned char type, unsigned char duration);
    int accessControlKnobCommand( unsigned char deviceAddress, unsigned char knobCommand, unsigned short masterId, unsigned char knobId, const char *params, unsigned char *data, int *sizeData );

    int customCommand( unsigned char deviceAddress, unsigned int idendification, unsigned int customCommand, const char *dataIn, int sizeIn, char *dataOut, int *sizeOut );

    int startTransparencyMode( unsigned char deviceAddress, unsigned short id, unsigned short flags, unsigned int sizeReadBuffer, unsigned int sizeWriteBuffer, int timeout, int *errorNumber );
    int stopTransparencyMode( unsigned char deviceAddress, int *errorNumber );
    int isTransparencyMode();

    int ping( unsigned char deviceAddress, int attempts, int *errorNumber );

public:
    controller_segment::block_transfer_state getBlockTransferState(unsigned char deviceAddress)
    {
        controller_segment &controllerSegment = channel4_.get_controller_segment(deviceAddress);
        return controllerSegment.transfer_state;
    }

    int getBlockTransferActiveCount(controller_segment::block_transfer_type blockType)
    {
        return channel4_.get_controller_segment_active_count(blockType);
    }

    controller_segment::block_transfer_type getBlockTransferType(unsigned char deviceAddress)
    {
        controller_segment &controllerSegment = channel4_.get_controller_segment(deviceAddress);
        return controllerSegment.transfer_type;
    }

    void setBlockTransferSettings(unsigned char deviceAddress, unsigned int typesAllowed, unsigned short duration)
    {
        controller_segment &controllerSegment = channel4_.get_controller_segment(deviceAddress);
        controllerSegment.allowed_transfer_types = typesAllowed;
        if (typesAllowed != controller_segment::btt_none)
            controllerSegment.duration = std::chrono::seconds(duration);
        else
            controllerSegment.duration = std::chrono::seconds(0);

        if (deviceAddress != 255)
        {
            // Ggf. Flag fuer die uebernahme vom Global-Objekt zuruecksetzen.
            controllerSegment.transfer_flags &= ~controller_segment::cf_settings_from_global;
        }
        else
        {
            // Wird Global-Objekt auf 0 gesetzt muessen alle die dieses Flag bekommen haben auch zuruecksetzen.
            channel4_.reset_controller_segment_global_duration();
        }
    }

    void getBlockTransferSettings(unsigned char deviceAddress, unsigned int *typesAllowed, unsigned short *duration)
    {
        controller_segment &controllerSegment = channel4_.get_controller_segment(deviceAddress);
        *typesAllowed = controllerSegment.allowed_transfer_types;
        *duration = static_cast<unsigned short>(controllerSegment.duration.count());
    }

    int blockTransferResume(unsigned char deviceAddress)
    {
        int res = IV_ERROR_SUCCESS;
        controller_segment &controllerSegment = channel4_.get_controller_segment(deviceAddress);

        // Ist Transfer aktiv?
        if (controllerSegment.transfer_state != controller_segment::bts_active)
            return -IV_ERROR_BLOCK_TRNASFER_INACTIVE;

        switch (controllerSegment.transfer_type)
        {
        case controller_segment::btt_list:
            {
                int errorNumber = IV_ERROR_SUCCESS;
                res = writeListData(deviceAddress, &errorNumber);
                if (res == 0)
                    res = -errorNumber; // Fehlerwert zurueckgeben.

                break;
            }

        case controller_segment::btt_access_control_list:
            {
                int errorNumber = IV_ERROR_SUCCESS;
                res = writeEntrance2ListData(deviceAddress, -1, &errorNumber);
                if (res == 0)
                    res = -errorNumber; // Fehlerwert zurueckgeben.

                break;
            }

        case controller_segment::btt_write_file:
            {
                res = writeFileData(deviceAddress, std::string());
                break;
            }

        case controller_segment::btt_read_file:
            {
                res = readFileData(deviceAddress, std::string());
                break;
            }

        case controller_segment::btt_write_setup_file:
            {
                res = channel4_.writeSetupData(deviceAddress, std::string());
                break;
            }

        case controller_segment::btt_read_setup_file:
            {
                res = channel4_.readSetupData(deviceAddress, std::string());
                break;
            }
        }

        return res;
    }

    int blockTransferDiscard(unsigned char deviceAddress)
    {
        controller_segment &controllerSegment = channel4_.get_controller_segment(deviceAddress);

        // Ist Transfer aktiv?
        if (controllerSegment.transfer_state != controller_segment::bts_active)
            return -IV_ERROR_BLOCK_TRNASFER_INACTIVE;

        if (isOpen())
            blockTransferResume(deviceAddress);

        // Sicherheitshalber noch Inactiv setzen, falls Kommunikationsfehler oder Schnittstelle geschlossen.
        controllerSegment.transfer_state = controller_segment::bts_inactive;

        return IV_ERROR_SUCCESS;
    }
};

#endif // !defined(AFX_AE3CHANNEL_H__D822D71E_AB5D_4A01_8121_C09FDB26DE7D__INCLUDED_)

